/*
 *  popd - a lean and hopefully secure pop3 implementation for
 *  maildir folders.
 * 
 *  Copyright (C) 1998  Thomas Roessler <roessler@guug.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <syslog.h>

#include "helpers.h"
#include "popd.h"

enum cmd {
  C_USER,
  C_PASS,
  C_STAT,
  C_LIST,
  C_RETR,
  C_DELE,
  C_NOOP,
  C_RSET,
  C_QUIT,
  C_NONE
};
    

static struct cmds
{
  enum cmd code;
  const char *str;
  int statemask;
  int flags;
} commands[] = {
  { C_USER, "USER", S_AUTH, 			CF_ARG },
  { C_PASS, "PASS", S_AUTHPASS, 		CF_ARG | CF_NOLOGARG},
  { C_STAT, "STAT", S_TRANS, 			0 },
  { C_LIST, "LIST", S_TRANS, 			CF_MAYBE },
  { C_RETR, "RETR", S_TRANS, 			CF_ARG },
  { C_DELE, "DELE", S_TRANS,			CF_ARG },
  { C_NOOP, "NOOP", S_TRANS | S_AUTH,		0 },
  { C_RSET, "RSET", S_TRANS,			0 },
  { C_QUIT, "QUIT", S_TRANS | S_AUTH, 		0 },
  { C_NONE, NULL,   0, 				0}
};


static enum cmd read_cmd(char **arg, struct context *ctx)
{
  char *s = NULL;
  char *t;
  char *logstr;
  int i;
  struct cmds *what = NULL;

  *arg = NULL;
  
  do {
    s = pop_getln();
    
    if(wannaquit) break;
    
    if(!s)
    {
      pop_err("Unknown command.");
      continue;
    }

    logstr = safe_strdup(s);

    if(!*s || !(t = strtok(s, " ")))
    {
      pop_err("Unkown command.");
      safe_free((void **)&s);
      safe_free((void **)&logstr);
      continue;
    }

    for(i = 0; commands[i].str && !what; i++)
    {
      if(!strcasecmp(t, commands[i].str))
	what = &commands[i];
    }
    
    if(!what)
    {
      pop_err("Unknown command.");
    }
    
    if(what && !(what->statemask & ctx->state))
    {
      pop_err("Command invalid.");
      what = NULL;
    }
    
    if(what && what->flags & (CF_ARG | CF_MAYBE))
    {
      t = strtok(NULL, " ");
      if(!t)
      {
	if(what->flags & CF_ARG)
	{
	  pop_err("Command requires an argument.");
	  what = NULL;
	}
	else
	  *arg = NULL;
      }
      else
	*arg = safe_strdup(t);
    }

    if(what)
    {
      if(*arg)
	sprintf(logstr, "%s %s", what->str, what->flags & CF_NOLOGARG ? "" : *arg);
      else
	sprintf(logstr, "%s", what->str);
      
      pop_log(logstr);
      
    }

    safe_free((void **)&logstr);
    safe_free((void **)&s);

  } while(!what && !wannaquit); 
  
  return what ? what->code : C_NONE;
}

int main(int argc, const char *argv[])
{
  struct context ctx;
  char *arg = NULL;

  wannaquit = 0;
  
  ctx.state = S_AUTHUSER;
  ctx.mailbox = NULL;
  ctx.p_msgs = 0;
  ctx.n_msgs = 0;
  ctx.s_mbx = 0;
  ctx.badauth = 0;
  ctx.authinfo = 0;
  ctx.what = NULL;

  openlog(argv[0], LOG_PID, LOG_DAEMON);
  
  pop_ok("POP3 server ready");
  
  while(ctx.state != S_QUIT && !wannaquit)
  {
    switch(read_cmd(&arg, &ctx))
    {
      case C_USER: do_user(&ctx, arg); ctx.state = S_AUTHPASS; break;
      case C_PASS: if(do_pass(&ctx, arg) == 0) ctx.state = S_TRANS; break;
      case C_STAT: do_stat(&ctx); break;
      case C_LIST: do_list(&ctx, arg); break;
      case C_RETR: do_retr(&ctx, arg); break;
      case C_DELE: do_dele(&ctx, arg); break;
      case C_NOOP: pop_ok(""); break;
      case C_RSET: do_rset(&ctx); break;
      
      case C_QUIT: 
      	if(ctx.state == S_TRANS)
	  do_update(&ctx);
        pop_ok("leaving");
        ctx.state = S_QUIT;
        break;

      case C_NONE: break;

      default:     pop_err("Command not implemented."); break;
    }

    if(arg) safe_free((void **)&arg);
    
  }

  return 0;
}
